# 实验5 实验报告

## 题目1 口算RSA的神童

口算是不可能口算的，可以使用 yafu 计算分解。

根据 RSA 的加密方式，先写好程序等待备用：

```python
import gmpy2
N = gmpy2.mpz(xxxxx)
p = gmpy2.mpz(xxxxx)
q = gmpy2.mpz(xxxxx)
c = gmpy2.mpz(xxxxx)
L = gmpy2.lcm(p-1, q-1)
e = 65537
D = gmpy2.invert(e, L)
ret = gmpy2.powmol(c, D, N)
```

ret 值就是题目的解。

下载 yapu，输入

```text
factor(${N})
```

啪的一声，很快嗷，就算出来了。

填到上边的程序里就能解出来了。

PS: yapu 在 linux 上好像不怎么方便编译

## 题目2 Agent008

扩展欧几里得算法：

```python
import gmpy2
print(gmpy2.gcdext(17, 65537))
```

输出：(mpz(1), mpz(30841), mpz(-8))

可见：30841\*17+(-8)\*65537=1

在 mod n 条件下， $c_1=m^{65537}$ 而 $c_2=m^{17}$，因此 $m = c_1^{-8}\times c_2^{30841}$

按照这个思路：

```python
import gmpy2

n = gmpy2.mpz(xxxxx)
c_1 = gmpy2.mpz(xxxxx)
c_2 = gmpy2.mpz(xxxxx)

c01 = gmpy2.powmod(c_1, gmpy2.mpz(-8), n)
c02 = gmpy2.powmod(c_2, gmpy2.mpz(30841), n)
m = gmpy2.powmod(c01 * c02, gmpy2.mpz(1), n)
print(m)
```

其中把 xxxxx 全都换成自己的数字，得到的结果填回到 shell 里就能得到 flag

## 题目3 太弱了?太弱了?

根据题目知：passed < 512n 且 c 是 passwd 的立方模 n，换言之知道 c+kn 如果是一个数的三次方，那么就可以知道这个数字是 passwd

显然如果存在，则 k 必然小于 512，这样看来是可以遍历的。也就是说：

```C
for k=0 to 512
   if ((c+nk)^{1/3} is a int)print((c+nk)^{1/3})
```

考虑到 gmpy2 的特性，编写代码如下：

```python
import gmpy2

n = gmpy2.mpz(xxx)
c = gmpy2.mpz(xxx)

for k in range(512):
    _c = c + k * n
    _c_3 = gmpy2.iroot(_c, 3)
    if _c_3[1] is True:
        print(_c_3)
```

得到的数字即为 passwd

## 题目4 ddl源赖氏佐田

此题可以根据中国剩余定理推出 m^3，进而使用 iroot 得到 m 的值。

```python
def crt(c: list, n: list):
    # 乘积
    N = n[0] * n[1] * n[2]
    # 求N/ni
    N_ni = [n[1] * n[2], n[0] * n[2], n[0] * n[1]]
    # 求 N/n_i -1 mod n_i
    N_ni_ = []
    for i in range(3):
        N_ni_.append(gmpy2.powmod(N_ni[i], -1, n[i]))
    # 求MiM'ibi的累加
    y = 0
    for i in range(3):
        y += (N_ni[i] * N_ni_[i] * c[i])
    y = gmpy2.mod(y, N)
    return y
```

把 C 和 N 全都放进去就行了。

PS: 我试了四次，最后的结果都是 false，我以为我算法有问题，检查了大约一个小时，最后发现纯属运气太好(差)。

## 题目5 自己发明一个RSA吧\!

先说一下这道题我一路做过来的想法：（因为这几天一直在外边跑，所以没有好好想，而且初步调试代码都是在我本地调试的）

首先看了讲义的 CSDN 链接，反正很多证明也没记住，但是对比题目源码可以得出：

- 首先，init 过程相比于 paillier 算法，没有检验 G 的准确性，但是经过我的简单手算，认为随机选择的 G 应该问题不大。
- 其次，加密过程中首先进行了一个 $\text{msg}^e\ \text{mod}\ N$ 的计算得到了一个变量 cip_rsa，**注意我此时还没有发现这个变量名中的 rsa 是什么意思**。以下定义“函数 f”为上边那个算式。
- 最后，解密过程我没看懂，但是我把 history 去掉之后发现确实能正确解密。

于是我就有一个想法：利用 **密文积=明文和**，采取选择明文攻击，应该说是随机明文攻击，就是随机找一个明文 rand 进行加密。同时由于我可以从代码知道 E 和 N，我可以把明文自己算一个 f(rand)，然后加密获得 e(rand)，同时我知道 e(pwd)，根据 e(pwd)\*e(rand)=f(rand)+f(pwd) 可以求出 f(pwd)，这样进行逆变换就能求解。

```python
import os

RTP_PWD2 = int.from_bytes(os.urandom(32).hex().encode(), 'big')
print("pwd2: ", RTP_PWD2)
N = int(input("N: "))
G = int(input("G: "))
E = 0x10001
RTP_PWD2_RSA = pow(RTP_PWD2, E, N)
RTP_PWD_E = int(input("pwd e: "))
RTP_PWD2_E = int(input("pwd2 e: "))
RTP_PWD_RSA = (RTP_PWD_E * RTP_PWD2_E - RTP_PWD2_RSA) % N
print(RTP_PWD_RSA)
print(pow(RTP_PWD_RSA, -E, N))
```

然后问题出在了逆变换上，经过几次失败我意识到 $f^{-1}(f(\text{pwd}))$ 并不好计算。此时我发现逆变换需要的指数只能是 RSA 的私钥，这显然是无法获得/破解的，所以 $f^{-1}$ 的方法行不通。

然后我注意到有一个关键点，就是加密虽然需要随机数 r ，但是解密和 r 无关。这样一来，我把密码的密文随便乘一个 $r_{new}^N(\text{mod}N^2)$ ，再塞回去仍然应该符合密文要求，并且就能规避历史记录。其中 $r_{new}$ 是我随机生成的数字。于是先试试：

```python
N = int(input("N: "))
cip = int(input("cip: "))
r_n = pow(randint(5, 500), N, N * N)
new_cip = (cip * r_n) % (N * N)
print(new_cip)
```

把输出结果去解密就可以得到密码（如果运气不太差），然后提交即可。
